// Autogenerated code; DO NOT EDIT.

/*
 * Schema Open API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type VirtualMachineProcessor struct {
	Count       uint32 `json:"Count,omitempty"`
	Limit       uint64 `json:"Limit,omitempty"`
	Weight      uint64 `json:"Weight,omitempty"`
	Reservation uint64 `json:"Reservation,omitempty"`
	CpuGroup                   *CpuGroup            `json:"CpuGroup,omitempty"`
	NumaProcessorsSettings     *NumaProcessors      `json:"NumaProcessorsSettings,omitempty"`
}
