// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.6
// 	protoc        v5.26.0
// source: github.com/Microsoft/hcsshim/cmd/containerd-shim-runhcs-v1/options/runhcs.proto

package options

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	timestamppb "google.golang.org/protobuf/types/known/timestamppb"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type Options_DebugType int32

const (
	Options_NPIPE Options_DebugType = 0
	Options_FILE  Options_DebugType = 1
	Options_ETW   Options_DebugType = 2
)

// Enum value maps for Options_DebugType.
var (
	Options_DebugType_name = map[int32]string{
		0: "NPIPE",
		1: "FILE",
		2: "ETW",
	}
	Options_DebugType_value = map[string]int32{
		"NPIPE": 0,
		"FILE":  1,
		"ETW":   2,
	}
)

func (x Options_DebugType) Enum() *Options_DebugType {
	p := new(Options_DebugType)
	*p = x
	return p
}

func (x Options_DebugType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (Options_DebugType) Descriptor() protoreflect.EnumDescriptor {
	return file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_enumTypes[0].Descriptor()
}

func (Options_DebugType) Type() protoreflect.EnumType {
	return &file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_enumTypes[0]
}

func (x Options_DebugType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use Options_DebugType.Descriptor instead.
func (Options_DebugType) EnumDescriptor() ([]byte, []int) {
	return file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescGZIP(), []int{0, 0}
}

type Options_SandboxIsolation int32

const (
	Options_PROCESS    Options_SandboxIsolation = 0
	Options_HYPERVISOR Options_SandboxIsolation = 1
)

// Enum value maps for Options_SandboxIsolation.
var (
	Options_SandboxIsolation_name = map[int32]string{
		0: "PROCESS",
		1: "HYPERVISOR",
	}
	Options_SandboxIsolation_value = map[string]int32{
		"PROCESS":    0,
		"HYPERVISOR": 1,
	}
)

func (x Options_SandboxIsolation) Enum() *Options_SandboxIsolation {
	p := new(Options_SandboxIsolation)
	*p = x
	return p
}

func (x Options_SandboxIsolation) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (Options_SandboxIsolation) Descriptor() protoreflect.EnumDescriptor {
	return file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_enumTypes[1].Descriptor()
}

func (Options_SandboxIsolation) Type() protoreflect.EnumType {
	return &file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_enumTypes[1]
}

func (x Options_SandboxIsolation) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use Options_SandboxIsolation.Descriptor instead.
func (Options_SandboxIsolation) EnumDescriptor() ([]byte, []int) {
	return file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescGZIP(), []int{0, 1}
}

// Options are the set of customizations that can be passed at Create time.
type Options struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Enable debug tracing (sets the logrus log level to debug). This may be deprecated in the future, prefer
	// log_level as this will override debug if both of them are set.
	Debug bool `protobuf:"varint,1,opt,name=debug,proto3" json:"debug,omitempty"`
	// debug tracing output type
	DebugType Options_DebugType `protobuf:"varint,2,opt,name=debug_type,json=debugType,proto3,enum=containerd.runhcs.v1.Options_DebugType" json:"debug_type,omitempty"`
	// registry key root for storage of the runhcs container state
	RegistryRoot string `protobuf:"bytes,3,opt,name=registry_root,json=registryRoot,proto3" json:"registry_root,omitempty"`
	// sandbox_image is the image to use for the sandbox that matches the
	// sandbox_platform.
	SandboxImage string `protobuf:"bytes,4,opt,name=sandbox_image,json=sandboxImage,proto3" json:"sandbox_image,omitempty"`
	// sandbox_platform is a CRI setting that specifies the platform
	// architecture for all sandbox's in this runtime. Values are
	// 'windows/amd64' and 'linux/amd64'.
	SandboxPlatform string `protobuf:"bytes,5,opt,name=sandbox_platform,json=sandboxPlatform,proto3" json:"sandbox_platform,omitempty"`
	// sandbox_isolation is a CRI setting that specifies the isolation level of
	// the sandbox. For Windows runtime PROCESS and HYPERVISOR are valid. For
	// LCOW only HYPERVISOR is valid and default if omitted.
	SandboxIsolation Options_SandboxIsolation `protobuf:"varint,6,opt,name=sandbox_isolation,json=sandboxIsolation,proto3,enum=containerd.runhcs.v1.Options_SandboxIsolation" json:"sandbox_isolation,omitempty"`
	// boot_files_root_path is the path to the directory containing the LCOW
	// kernel and root FS files.
	BootFilesRootPath string `protobuf:"bytes,7,opt,name=boot_files_root_path,json=bootFilesRootPath,proto3" json:"boot_files_root_path,omitempty"`
	// vm_processor_count is the default number of processors to create for the
	// hypervisor isolated utility vm.
	//
	// The platform default if omitted is 2, unless the host only has a single
	// core in which case it is 1.
	VmProcessorCount int32 `protobuf:"varint,8,opt,name=vm_processor_count,json=vmProcessorCount,proto3" json:"vm_processor_count,omitempty"`
	// vm_memory_size_in_mb is the default amount of memory to assign to the
	// hypervisor isolated utility vm.
	//
	// The platform default is 1024MB if omitted.
	VmMemorySizeInMb int32 `protobuf:"varint,9,opt,name=vm_memory_size_in_mb,json=vmMemorySizeInMb,proto3" json:"vm_memory_size_in_mb,omitempty"`
	// GPUVHDPath is the path to the gpu vhd to add to the uvm
	// when a container requests a gpu
	GPUVHDPath string `protobuf:"bytes,10,opt,name=GPUVHDPath,proto3" json:"GPUVHDPath,omitempty"`
	// scale_cpu_limits_to_sandbox indicates that container CPU limits should
	// be adjusted to account for the difference in number of cores between the
	// host and UVM.
	ScaleCpuLimitsToSandbox bool `protobuf:"varint,11,opt,name=scale_cpu_limits_to_sandbox,json=scaleCpuLimitsToSandbox,proto3" json:"scale_cpu_limits_to_sandbox,omitempty"`
	// default_container_scratch_size_in_gb is the default scratch size (sandbox.vhdx)
	// to be used for containers. Every container will get a sandbox of `size_in_gb` assigned
	// instead of the default of 20GB.
	DefaultContainerScratchSizeInGb int32 `protobuf:"varint,12,opt,name=default_container_scratch_size_in_gb,json=defaultContainerScratchSizeInGb,proto3" json:"default_container_scratch_size_in_gb,omitempty"`
	// default_vm_scratch_size_in_gb is the default scratch size (sandbox.vhdx)
	// to be used for the UVM. This only applies to WCOW as LCOW doesn't mount a scratch
	// specifically for the UVM.
	DefaultVmScratchSizeInGb int32 `protobuf:"varint,13,opt,name=default_vm_scratch_size_in_gb,json=defaultVmScratchSizeInGb,proto3" json:"default_vm_scratch_size_in_gb,omitempty"`
	// share_scratch specifies if we'd like to reuse scratch space between multiple containers.
	// This currently only affects LCOW. The sandbox containers scratch space is re-used for all
	// subsequent containers launched in the pod.
	ShareScratch bool `protobuf:"varint,14,opt,name=share_scratch,json=shareScratch,proto3" json:"share_scratch,omitempty"`
	// NCProxyAddr is the address of the network configuration proxy service. If omitted
	// the network is setup locally.
	NCProxyAddr string `protobuf:"bytes,15,opt,name=NCProxyAddr,proto3" json:"NCProxyAddr,omitempty"`
	// log_level specifies the logrus log level for the shim. Supported values are a string representation of the
	// logrus log levels: "trace", "debug", "info", "warn", "error", "fatal", "panic". This setting will override
	// the `debug` field if both are specified, unless the level specified is also "debug", as these are equivalent.
	LogLevel string `protobuf:"bytes,16,opt,name=log_level,json=logLevel,proto3" json:"log_level,omitempty"`
	// io_retry_timeout_in_sec is the timeout in seconds for how long to try and reconnect to an upstream IO provider if a connection is lost.
	// The typical example is if Containerd has restarted but is expected to come back online. A 0 for this field is interpreted as an infinite
	// timeout.
	IoRetryTimeoutInSec int32 `protobuf:"varint,17,opt,name=io_retry_timeout_in_sec,json=ioRetryTimeoutInSec,proto3" json:"io_retry_timeout_in_sec,omitempty"`
	// default_container_annotations specifies a set of annotations that should be set for every workload container
	DefaultContainerAnnotations map[string]string `protobuf:"bytes,18,rep,name=default_container_annotations,json=defaultContainerAnnotations,proto3" json:"default_container_annotations,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	// no_inherit_host_timezone specifies to skip inheriting the hosts time zone for WCOW UVMs and instead default to
	// UTC.
	NoInheritHostTimezone bool `protobuf:"varint,19,opt,name=no_inherit_host_timezone,json=noInheritHostTimezone,proto3" json:"no_inherit_host_timezone,omitempty"`
	// scrub_logs enables removing environment variables and other potentially sensitive information from logs
	ScrubLogs     bool `protobuf:"varint,20,opt,name=scrub_logs,json=scrubLogs,proto3" json:"scrub_logs,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Options) Reset() {
	*x = Options{}
	mi := &file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Options) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Options) ProtoMessage() {}

func (x *Options) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Options.ProtoReflect.Descriptor instead.
func (*Options) Descriptor() ([]byte, []int) {
	return file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescGZIP(), []int{0}
}

func (x *Options) GetDebug() bool {
	if x != nil {
		return x.Debug
	}
	return false
}

func (x *Options) GetDebugType() Options_DebugType {
	if x != nil {
		return x.DebugType
	}
	return Options_NPIPE
}

func (x *Options) GetRegistryRoot() string {
	if x != nil {
		return x.RegistryRoot
	}
	return ""
}

func (x *Options) GetSandboxImage() string {
	if x != nil {
		return x.SandboxImage
	}
	return ""
}

func (x *Options) GetSandboxPlatform() string {
	if x != nil {
		return x.SandboxPlatform
	}
	return ""
}

func (x *Options) GetSandboxIsolation() Options_SandboxIsolation {
	if x != nil {
		return x.SandboxIsolation
	}
	return Options_PROCESS
}

func (x *Options) GetBootFilesRootPath() string {
	if x != nil {
		return x.BootFilesRootPath
	}
	return ""
}

func (x *Options) GetVmProcessorCount() int32 {
	if x != nil {
		return x.VmProcessorCount
	}
	return 0
}

func (x *Options) GetVmMemorySizeInMb() int32 {
	if x != nil {
		return x.VmMemorySizeInMb
	}
	return 0
}

func (x *Options) GetGPUVHDPath() string {
	if x != nil {
		return x.GPUVHDPath
	}
	return ""
}

func (x *Options) GetScaleCpuLimitsToSandbox() bool {
	if x != nil {
		return x.ScaleCpuLimitsToSandbox
	}
	return false
}

func (x *Options) GetDefaultContainerScratchSizeInGb() int32 {
	if x != nil {
		return x.DefaultContainerScratchSizeInGb
	}
	return 0
}

func (x *Options) GetDefaultVmScratchSizeInGb() int32 {
	if x != nil {
		return x.DefaultVmScratchSizeInGb
	}
	return 0
}

func (x *Options) GetShareScratch() bool {
	if x != nil {
		return x.ShareScratch
	}
	return false
}

func (x *Options) GetNCProxyAddr() string {
	if x != nil {
		return x.NCProxyAddr
	}
	return ""
}

func (x *Options) GetLogLevel() string {
	if x != nil {
		return x.LogLevel
	}
	return ""
}

func (x *Options) GetIoRetryTimeoutInSec() int32 {
	if x != nil {
		return x.IoRetryTimeoutInSec
	}
	return 0
}

func (x *Options) GetDefaultContainerAnnotations() map[string]string {
	if x != nil {
		return x.DefaultContainerAnnotations
	}
	return nil
}

func (x *Options) GetNoInheritHostTimezone() bool {
	if x != nil {
		return x.NoInheritHostTimezone
	}
	return false
}

func (x *Options) GetScrubLogs() bool {
	if x != nil {
		return x.ScrubLogs
	}
	return false
}

// ProcessDetails contains additional information about a process. This is the additional
// info returned in the Pids query.
type ProcessDetails struct {
	state                        protoimpl.MessageState `protogen:"open.v1"`
	ImageName                    string                 `protobuf:"bytes,1,opt,name=image_name,json=imageName,proto3" json:"image_name,omitempty"`
	CreatedAt                    *timestamppb.Timestamp `protobuf:"bytes,2,opt,name=created_at,json=createdAt,proto3" json:"created_at,omitempty"`
	KernelTime_100Ns             uint64                 `protobuf:"varint,3,opt,name=kernel_time_100_ns,json=kernelTime100Ns,proto3" json:"kernel_time_100_ns,omitempty"`
	MemoryCommitBytes            uint64                 `protobuf:"varint,4,opt,name=memory_commit_bytes,json=memoryCommitBytes,proto3" json:"memory_commit_bytes,omitempty"`
	MemoryWorkingSetPrivateBytes uint64                 `protobuf:"varint,5,opt,name=memory_working_set_private_bytes,json=memoryWorkingSetPrivateBytes,proto3" json:"memory_working_set_private_bytes,omitempty"`
	MemoryWorkingSetSharedBytes  uint64                 `protobuf:"varint,6,opt,name=memory_working_set_shared_bytes,json=memoryWorkingSetSharedBytes,proto3" json:"memory_working_set_shared_bytes,omitempty"`
	ProcessID                    uint32                 `protobuf:"varint,7,opt,name=process_id,json=processId,proto3" json:"process_id,omitempty"`
	UserTime_100Ns               uint64                 `protobuf:"varint,8,opt,name=user_time_100_ns,json=userTime100Ns,proto3" json:"user_time_100_ns,omitempty"`
	ExecID                       string                 `protobuf:"bytes,9,opt,name=exec_id,json=execId,proto3" json:"exec_id,omitempty"`
	unknownFields                protoimpl.UnknownFields
	sizeCache                    protoimpl.SizeCache
}

func (x *ProcessDetails) Reset() {
	*x = ProcessDetails{}
	mi := &file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProcessDetails) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProcessDetails) ProtoMessage() {}

func (x *ProcessDetails) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProcessDetails.ProtoReflect.Descriptor instead.
func (*ProcessDetails) Descriptor() ([]byte, []int) {
	return file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescGZIP(), []int{1}
}

func (x *ProcessDetails) GetImageName() string {
	if x != nil {
		return x.ImageName
	}
	return ""
}

func (x *ProcessDetails) GetCreatedAt() *timestamppb.Timestamp {
	if x != nil {
		return x.CreatedAt
	}
	return nil
}

func (x *ProcessDetails) GetKernelTime_100Ns() uint64 {
	if x != nil {
		return x.KernelTime_100Ns
	}
	return 0
}

func (x *ProcessDetails) GetMemoryCommitBytes() uint64 {
	if x != nil {
		return x.MemoryCommitBytes
	}
	return 0
}

func (x *ProcessDetails) GetMemoryWorkingSetPrivateBytes() uint64 {
	if x != nil {
		return x.MemoryWorkingSetPrivateBytes
	}
	return 0
}

func (x *ProcessDetails) GetMemoryWorkingSetSharedBytes() uint64 {
	if x != nil {
		return x.MemoryWorkingSetSharedBytes
	}
	return 0
}

func (x *ProcessDetails) GetProcessID() uint32 {
	if x != nil {
		return x.ProcessID
	}
	return 0
}

func (x *ProcessDetails) GetUserTime_100Ns() uint64 {
	if x != nil {
		return x.UserTime_100Ns
	}
	return 0
}

func (x *ProcessDetails) GetExecID() string {
	if x != nil {
		return x.ExecID
	}
	return ""
}

var File_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto protoreflect.FileDescriptor

const file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDesc = "" +
	"\n" +
	"Ogithub.com/Microsoft/hcsshim/cmd/containerd-shim-runhcs-v1/options/runhcs.proto\x12\x14containerd.runhcs.v1\x1a\x1fgoogle/protobuf/timestamp.proto\"\xd9\t\n" +
	"\aOptions\x12\x14\n" +
	"\x05debug\x18\x01 \x01(\bR\x05debug\x12F\n" +
	"\n" +
	"debug_type\x18\x02 \x01(\x0e2'.containerd.runhcs.v1.Options.DebugTypeR\tdebugType\x12#\n" +
	"\rregistry_root\x18\x03 \x01(\tR\fregistryRoot\x12#\n" +
	"\rsandbox_image\x18\x04 \x01(\tR\fsandboxImage\x12)\n" +
	"\x10sandbox_platform\x18\x05 \x01(\tR\x0fsandboxPlatform\x12[\n" +
	"\x11sandbox_isolation\x18\x06 \x01(\x0e2..containerd.runhcs.v1.Options.SandboxIsolationR\x10sandboxIsolation\x12/\n" +
	"\x14boot_files_root_path\x18\a \x01(\tR\x11bootFilesRootPath\x12,\n" +
	"\x12vm_processor_count\x18\b \x01(\x05R\x10vmProcessorCount\x12.\n" +
	"\x14vm_memory_size_in_mb\x18\t \x01(\x05R\x10vmMemorySizeInMb\x12\x1e\n" +
	"\n" +
	"GPUVHDPath\x18\n" +
	" \x01(\tR\n" +
	"GPUVHDPath\x12<\n" +
	"\x1bscale_cpu_limits_to_sandbox\x18\v \x01(\bR\x17scaleCpuLimitsToSandbox\x12M\n" +
	"$default_container_scratch_size_in_gb\x18\f \x01(\x05R\x1fdefaultContainerScratchSizeInGb\x12?\n" +
	"\x1ddefault_vm_scratch_size_in_gb\x18\r \x01(\x05R\x18defaultVmScratchSizeInGb\x12#\n" +
	"\rshare_scratch\x18\x0e \x01(\bR\fshareScratch\x12 \n" +
	"\vNCProxyAddr\x18\x0f \x01(\tR\vNCProxyAddr\x12\x1b\n" +
	"\tlog_level\x18\x10 \x01(\tR\blogLevel\x124\n" +
	"\x17io_retry_timeout_in_sec\x18\x11 \x01(\x05R\x13ioRetryTimeoutInSec\x12\x82\x01\n" +
	"\x1ddefault_container_annotations\x18\x12 \x03(\v2>.containerd.runhcs.v1.Options.DefaultContainerAnnotationsEntryR\x1bdefaultContainerAnnotations\x127\n" +
	"\x18no_inherit_host_timezone\x18\x13 \x01(\bR\x15noInheritHostTimezone\x12\x1d\n" +
	"\n" +
	"scrub_logs\x18\x14 \x01(\bR\tscrubLogs\x1aN\n" +
	" DefaultContainerAnnotationsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\")\n" +
	"\tDebugType\x12\t\n" +
	"\x05NPIPE\x10\x00\x12\b\n" +
	"\x04FILE\x10\x01\x12\a\n" +
	"\x03ETW\x10\x02\"/\n" +
	"\x10SandboxIsolation\x12\v\n" +
	"\aPROCESS\x10\x00\x12\x0e\n" +
	"\n" +
	"HYPERVISOR\x10\x01\"\xb6\x03\n" +
	"\x0eProcessDetails\x12\x1d\n" +
	"\n" +
	"image_name\x18\x01 \x01(\tR\timageName\x129\n" +
	"\n" +
	"created_at\x18\x02 \x01(\v2\x1a.google.protobuf.TimestampR\tcreatedAt\x12+\n" +
	"\x12kernel_time_100_ns\x18\x03 \x01(\x04R\x0fkernelTime100Ns\x12.\n" +
	"\x13memory_commit_bytes\x18\x04 \x01(\x04R\x11memoryCommitBytes\x12F\n" +
	" memory_working_set_private_bytes\x18\x05 \x01(\x04R\x1cmemoryWorkingSetPrivateBytes\x12D\n" +
	"\x1fmemory_working_set_shared_bytes\x18\x06 \x01(\x04R\x1bmemoryWorkingSetSharedBytes\x12\x1d\n" +
	"\n" +
	"process_id\x18\a \x01(\rR\tprocessId\x12'\n" +
	"\x10user_time_100_ns\x18\b \x01(\x04R\ruserTime100Ns\x12\x17\n" +
	"\aexec_id\x18\t \x01(\tR\x06execIdBDZBgithub.com/Microsoft/hcsshim/cmd/containerd-shim-runhcs-v1/optionsb\x06proto3"

var (
	file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescOnce sync.Once
	file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescData []byte
)

func file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescGZIP() []byte {
	file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescOnce.Do(func() {
		file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDesc), len(file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDesc)))
	})
	return file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDescData
}

var file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_enumTypes = make([]protoimpl.EnumInfo, 2)
var file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_msgTypes = make([]protoimpl.MessageInfo, 3)
var file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_goTypes = []any{
	(Options_DebugType)(0),        // 0: containerd.runhcs.v1.Options.DebugType
	(Options_SandboxIsolation)(0), // 1: containerd.runhcs.v1.Options.SandboxIsolation
	(*Options)(nil),               // 2: containerd.runhcs.v1.Options
	(*ProcessDetails)(nil),        // 3: containerd.runhcs.v1.ProcessDetails
	nil,                           // 4: containerd.runhcs.v1.Options.DefaultContainerAnnotationsEntry
	(*timestamppb.Timestamp)(nil), // 5: google.protobuf.Timestamp
}
var file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_depIdxs = []int32{
	0, // 0: containerd.runhcs.v1.Options.debug_type:type_name -> containerd.runhcs.v1.Options.DebugType
	1, // 1: containerd.runhcs.v1.Options.sandbox_isolation:type_name -> containerd.runhcs.v1.Options.SandboxIsolation
	4, // 2: containerd.runhcs.v1.Options.default_container_annotations:type_name -> containerd.runhcs.v1.Options.DefaultContainerAnnotationsEntry
	5, // 3: containerd.runhcs.v1.ProcessDetails.created_at:type_name -> google.protobuf.Timestamp
	4, // [4:4] is the sub-list for method output_type
	4, // [4:4] is the sub-list for method input_type
	4, // [4:4] is the sub-list for extension type_name
	4, // [4:4] is the sub-list for extension extendee
	0, // [0:4] is the sub-list for field type_name
}

func init() {
	file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_init()
}
func file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_init() {
	if File_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDesc), len(file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_rawDesc)),
			NumEnums:      2,
			NumMessages:   3,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_goTypes,
		DependencyIndexes: file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_depIdxs,
		EnumInfos:         file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_enumTypes,
		MessageInfos:      file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_msgTypes,
	}.Build()
	File_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto = out.File
	file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_goTypes = nil
	file_github_com_Microsoft_hcsshim_cmd_containerd_shim_runhcs_v1_options_runhcs_proto_depIdxs = nil
}
