/* Do not edit this file manually.
   This file is autogenerated by windmc.  */


// Country: United States
// Language: English
#pragma code_page(1252)
LANGUAGE 0x9, 0x1
1 MESSAGETABLE "MSG00001.bin"
