// Autogenerated code; DO NOT EDIT.

/*
 * Schema Open API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

// VirtualSlitType : Indicates if a virtual SLIT should ne enabled for a VM and the type of virtual SLIT to be enabled.
type VirtualSlitType string

// List of VirtualSlitType
const (
	VirtualSlitType_NONE                       VirtualSlitType = "None"
	VirtualSlitType_FIRMWARE                   VirtualSlitType = "Firmware"
	VirtualSlitType_MEASURED                   VirtualSlitType = "Measured"
	VirtualSlitType_FIRMWARE_FALLBACK_MEASURED VirtualSlitType = "FirmwareFallbackMeasured"
)
