/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.5
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type CimMount struct {
	ImagePath      string `json:"ImagePath,omitempty"`
	FileSystemName string `json:"FileSystemName,omitempty"`
	VolumeGuid     string `json:"VolumeGuid,omitempty"`
	MountFlags     uint32 `json:"MountFlags,omitempty"`
}
