// Copyright 2025 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.33.2
// source: google/logging/v2/logging.proto

package loggingpb

import (
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"

	_ "google.golang.org/genproto/googleapis/api/annotations"
	monitoredres "google.golang.org/genproto/googleapis/api/monitoredres"
	status "google.golang.org/genproto/googleapis/rpc/status"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	durationpb "google.golang.org/protobuf/types/known/durationpb"
	emptypb "google.golang.org/protobuf/types/known/emptypb"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// An indicator of why entries were omitted.
type TailLogEntriesResponse_SuppressionInfo_Reason int32

const (
	// Unexpected default.
	TailLogEntriesResponse_SuppressionInfo_REASON_UNSPECIFIED TailLogEntriesResponse_SuppressionInfo_Reason = 0
	// Indicates suppression occurred due to relevant entries being
	// received in excess of rate limits. For quotas and limits, see
	// [Logging API quotas and
	// limits](https://cloud.google.com/logging/quotas#api-limits).
	TailLogEntriesResponse_SuppressionInfo_RATE_LIMIT TailLogEntriesResponse_SuppressionInfo_Reason = 1
	// Indicates suppression occurred due to the client not consuming
	// responses quickly enough.
	TailLogEntriesResponse_SuppressionInfo_NOT_CONSUMED TailLogEntriesResponse_SuppressionInfo_Reason = 2
)

// Enum value maps for TailLogEntriesResponse_SuppressionInfo_Reason.
var (
	TailLogEntriesResponse_SuppressionInfo_Reason_name = map[int32]string{
		0: "REASON_UNSPECIFIED",
		1: "RATE_LIMIT",
		2: "NOT_CONSUMED",
	}
	TailLogEntriesResponse_SuppressionInfo_Reason_value = map[string]int32{
		"REASON_UNSPECIFIED": 0,
		"RATE_LIMIT":         1,
		"NOT_CONSUMED":       2,
	}
)

func (x TailLogEntriesResponse_SuppressionInfo_Reason) Enum() *TailLogEntriesResponse_SuppressionInfo_Reason {
	p := new(TailLogEntriesResponse_SuppressionInfo_Reason)
	*p = x
	return p
}

func (x TailLogEntriesResponse_SuppressionInfo_Reason) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (TailLogEntriesResponse_SuppressionInfo_Reason) Descriptor() protoreflect.EnumDescriptor {
	return file_google_logging_v2_logging_proto_enumTypes[0].Descriptor()
}

func (TailLogEntriesResponse_SuppressionInfo_Reason) Type() protoreflect.EnumType {
	return &file_google_logging_v2_logging_proto_enumTypes[0]
}

func (x TailLogEntriesResponse_SuppressionInfo_Reason) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use TailLogEntriesResponse_SuppressionInfo_Reason.Descriptor instead.
func (TailLogEntriesResponse_SuppressionInfo_Reason) EnumDescriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{11, 0, 0}
}

// The parameters to DeleteLog.
type DeleteLogRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Required. The resource name of the log to delete:
	//
	// * `projects/[PROJECT_ID]/logs/[LOG_ID]`
	// * `organizations/[ORGANIZATION_ID]/logs/[LOG_ID]`
	// * `billingAccounts/[BILLING_ACCOUNT_ID]/logs/[LOG_ID]`
	// * `folders/[FOLDER_ID]/logs/[LOG_ID]`
	//
	// `[LOG_ID]` must be URL-encoded. For example,
	// `"projects/my-project-id/logs/syslog"`,
	// `"organizations/123/logs/cloudaudit.googleapis.com%2Factivity"`.
	//
	// For more information about log names, see
	// [LogEntry][google.logging.v2.LogEntry].
	LogName       string `protobuf:"bytes,1,opt,name=log_name,json=logName,proto3" json:"log_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DeleteLogRequest) Reset() {
	*x = DeleteLogRequest{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DeleteLogRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DeleteLogRequest) ProtoMessage() {}

func (x *DeleteLogRequest) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DeleteLogRequest.ProtoReflect.Descriptor instead.
func (*DeleteLogRequest) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{0}
}

func (x *DeleteLogRequest) GetLogName() string {
	if x != nil {
		return x.LogName
	}
	return ""
}

// The parameters to WriteLogEntries.
type WriteLogEntriesRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Optional. A default log resource name that is assigned to all log entries
	// in `entries` that do not specify a value for `log_name`:
	//
	// * `projects/[PROJECT_ID]/logs/[LOG_ID]`
	// * `organizations/[ORGANIZATION_ID]/logs/[LOG_ID]`
	// * `billingAccounts/[BILLING_ACCOUNT_ID]/logs/[LOG_ID]`
	// * `folders/[FOLDER_ID]/logs/[LOG_ID]`
	//
	// `[LOG_ID]` must be URL-encoded. For example:
	//
	//	"projects/my-project-id/logs/syslog"
	//	"organizations/123/logs/cloudaudit.googleapis.com%2Factivity"
	//
	// The permission `logging.logEntries.create` is needed on each project,
	// organization, billing account, or folder that is receiving new log
	// entries, whether the resource is specified in `logName` or in an
	// individual log entry.
	LogName string `protobuf:"bytes,1,opt,name=log_name,json=logName,proto3" json:"log_name,omitempty"`
	// Optional. A default monitored resource object that is assigned to all log
	// entries in `entries` that do not specify a value for `resource`. Example:
	//
	//	{ "type": "gce_instance",
	//	  "labels": {
	//	    "zone": "us-central1-a", "instance_id": "00000000000000000000" }}
	//
	// See [LogEntry][google.logging.v2.LogEntry].
	Resource *monitoredres.MonitoredResource `protobuf:"bytes,2,opt,name=resource,proto3" json:"resource,omitempty"`
	// Optional. Default labels that are added to the `labels` field of all log
	// entries in `entries`. If a log entry already has a label with the same key
	// as a label in this parameter, then the log entry's label is not changed.
	// See [LogEntry][google.logging.v2.LogEntry].
	Labels map[string]string `protobuf:"bytes,3,rep,name=labels,proto3" json:"labels,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	// Required. The log entries to send to Logging. The order of log
	// entries in this list does not matter. Values supplied in this method's
	// `log_name`, `resource`, and `labels` fields are copied into those log
	// entries in this list that do not include values for their corresponding
	// fields. For more information, see the
	// [LogEntry][google.logging.v2.LogEntry] type.
	//
	// If the `timestamp` or `insert_id` fields are missing in log entries, then
	// this method supplies the current time or a unique identifier, respectively.
	// The supplied values are chosen so that, among the log entries that did not
	// supply their own values, the entries earlier in the list will sort before
	// the entries later in the list. See the `entries.list` method.
	//
	// Log entries with timestamps that are more than the
	// [logs retention period](https://cloud.google.com/logging/quotas) in
	// the past or more than 24 hours in the future will not be available when
	// calling `entries.list`. However, those log entries can still be [exported
	// with
	// LogSinks](https://cloud.google.com/logging/docs/api/tasks/exporting-logs).
	//
	// To improve throughput and to avoid exceeding the
	// [quota limit](https://cloud.google.com/logging/quotas) for calls to
	// `entries.write`, you should try to include several log entries in this
	// list, rather than calling this method for each individual log entry.
	Entries []*LogEntry `protobuf:"bytes,4,rep,name=entries,proto3" json:"entries,omitempty"`
	// Optional. Whether a batch's valid entries should be written even if some
	// other entry failed due to a permanent error such as INVALID_ARGUMENT or
	// PERMISSION_DENIED. If any entry failed, then the response status is the
	// response status of one of the failed entries. The response will include
	// error details in `WriteLogEntriesPartialErrors.log_entry_errors` keyed by
	// the entries' zero-based index in the `entries`. Failed requests for which
	// no entries are written will not include per-entry errors.
	PartialSuccess bool `protobuf:"varint,5,opt,name=partial_success,json=partialSuccess,proto3" json:"partial_success,omitempty"`
	// Optional. If true, the request should expect normal response, but the
	// entries won't be persisted nor exported. Useful for checking whether the
	// logging API endpoints are working properly before sending valuable data.
	DryRun        bool `protobuf:"varint,6,opt,name=dry_run,json=dryRun,proto3" json:"dry_run,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *WriteLogEntriesRequest) Reset() {
	*x = WriteLogEntriesRequest{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *WriteLogEntriesRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*WriteLogEntriesRequest) ProtoMessage() {}

func (x *WriteLogEntriesRequest) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use WriteLogEntriesRequest.ProtoReflect.Descriptor instead.
func (*WriteLogEntriesRequest) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{1}
}

func (x *WriteLogEntriesRequest) GetLogName() string {
	if x != nil {
		return x.LogName
	}
	return ""
}

func (x *WriteLogEntriesRequest) GetResource() *monitoredres.MonitoredResource {
	if x != nil {
		return x.Resource
	}
	return nil
}

func (x *WriteLogEntriesRequest) GetLabels() map[string]string {
	if x != nil {
		return x.Labels
	}
	return nil
}

func (x *WriteLogEntriesRequest) GetEntries() []*LogEntry {
	if x != nil {
		return x.Entries
	}
	return nil
}

func (x *WriteLogEntriesRequest) GetPartialSuccess() bool {
	if x != nil {
		return x.PartialSuccess
	}
	return false
}

func (x *WriteLogEntriesRequest) GetDryRun() bool {
	if x != nil {
		return x.DryRun
	}
	return false
}

// Result returned from WriteLogEntries.
type WriteLogEntriesResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *WriteLogEntriesResponse) Reset() {
	*x = WriteLogEntriesResponse{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *WriteLogEntriesResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*WriteLogEntriesResponse) ProtoMessage() {}

func (x *WriteLogEntriesResponse) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use WriteLogEntriesResponse.ProtoReflect.Descriptor instead.
func (*WriteLogEntriesResponse) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{2}
}

// Error details for WriteLogEntries with partial success.
type WriteLogEntriesPartialErrors struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// When `WriteLogEntriesRequest.partial_success` is true, records the error
	// status for entries that were not written due to a permanent error, keyed
	// by the entry's zero-based index in `WriteLogEntriesRequest.entries`.
	//
	// Failed requests for which no entries are written will not include
	// per-entry errors.
	LogEntryErrors map[int32]*status.Status `protobuf:"bytes,1,rep,name=log_entry_errors,json=logEntryErrors,proto3" json:"log_entry_errors,omitempty" protobuf_key:"varint,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields  protoimpl.UnknownFields
	sizeCache      protoimpl.SizeCache
}

func (x *WriteLogEntriesPartialErrors) Reset() {
	*x = WriteLogEntriesPartialErrors{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *WriteLogEntriesPartialErrors) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*WriteLogEntriesPartialErrors) ProtoMessage() {}

func (x *WriteLogEntriesPartialErrors) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use WriteLogEntriesPartialErrors.ProtoReflect.Descriptor instead.
func (*WriteLogEntriesPartialErrors) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{3}
}

func (x *WriteLogEntriesPartialErrors) GetLogEntryErrors() map[int32]*status.Status {
	if x != nil {
		return x.LogEntryErrors
	}
	return nil
}

// The parameters to `ListLogEntries`.
type ListLogEntriesRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Required. Names of one or more parent resources from which to
	// retrieve log entries:
	//
	// *  `projects/[PROJECT_ID]`
	// *  `organizations/[ORGANIZATION_ID]`
	// *  `billingAccounts/[BILLING_ACCOUNT_ID]`
	// *  `folders/[FOLDER_ID]`
	//
	// May alternatively be one or more views:
	//
	//   - `projects/[PROJECT_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `organizations/[ORGANIZATION_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `billingAccounts/[BILLING_ACCOUNT_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `folders/[FOLDER_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//
	// Projects listed in the `project_ids` field are added to this list.
	// A maximum of 100 resources may be specified in a single request.
	ResourceNames []string `protobuf:"bytes,8,rep,name=resource_names,json=resourceNames,proto3" json:"resource_names,omitempty"`
	// Optional. Only log entries that match the filter are returned.  An empty
	// filter matches all log entries in the resources listed in `resource_names`.
	// Referencing a parent resource that is not listed in `resource_names` will
	// cause the filter to return no results. The maximum length of a filter is
	// 20,000 characters.
	Filter string `protobuf:"bytes,2,opt,name=filter,proto3" json:"filter,omitempty"`
	// Optional. How the results should be sorted.  Presently, the only permitted
	// values are `"timestamp asc"` (default) and `"timestamp desc"`. The first
	// option returns entries in order of increasing values of
	// `LogEntry.timestamp` (oldest first), and the second option returns entries
	// in order of decreasing timestamps (newest first).  Entries with equal
	// timestamps are returned in order of their `insert_id` values.
	OrderBy string `protobuf:"bytes,3,opt,name=order_by,json=orderBy,proto3" json:"order_by,omitempty"`
	// Optional. The maximum number of results to return from this request.
	// Default is 50. If the value is negative or exceeds 1000, the request is
	// rejected. The presence of `next_page_token` in the response indicates that
	// more results might be available.
	PageSize int32 `protobuf:"varint,4,opt,name=page_size,json=pageSize,proto3" json:"page_size,omitempty"`
	// Optional. If present, then retrieve the next batch of results from the
	// preceding call to this method.  `page_token` must be the value of
	// `next_page_token` from the previous response.  The values of other method
	// parameters should be identical to those in the previous call.
	PageToken     string `protobuf:"bytes,5,opt,name=page_token,json=pageToken,proto3" json:"page_token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListLogEntriesRequest) Reset() {
	*x = ListLogEntriesRequest{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListLogEntriesRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListLogEntriesRequest) ProtoMessage() {}

func (x *ListLogEntriesRequest) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListLogEntriesRequest.ProtoReflect.Descriptor instead.
func (*ListLogEntriesRequest) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{4}
}

func (x *ListLogEntriesRequest) GetResourceNames() []string {
	if x != nil {
		return x.ResourceNames
	}
	return nil
}

func (x *ListLogEntriesRequest) GetFilter() string {
	if x != nil {
		return x.Filter
	}
	return ""
}

func (x *ListLogEntriesRequest) GetOrderBy() string {
	if x != nil {
		return x.OrderBy
	}
	return ""
}

func (x *ListLogEntriesRequest) GetPageSize() int32 {
	if x != nil {
		return x.PageSize
	}
	return 0
}

func (x *ListLogEntriesRequest) GetPageToken() string {
	if x != nil {
		return x.PageToken
	}
	return ""
}

// Result returned from `ListLogEntries`.
type ListLogEntriesResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// A list of log entries.  If `entries` is empty, `nextPageToken` may still be
	// returned, indicating that more entries may exist.  See `nextPageToken` for
	// more information.
	Entries []*LogEntry `protobuf:"bytes,1,rep,name=entries,proto3" json:"entries,omitempty"`
	// If there might be more results than those appearing in this response, then
	// `nextPageToken` is included.  To get the next set of results, call this
	// method again using the value of `nextPageToken` as `pageToken`.
	//
	// If a value for `next_page_token` appears and the `entries` field is empty,
	// it means that the search found no log entries so far but it did not have
	// time to search all the possible log entries.  Retry the method with this
	// value for `page_token` to continue the search.  Alternatively, consider
	// speeding up the search by changing your filter to specify a single log name
	// or resource type, or to narrow the time range of the search.
	NextPageToken string `protobuf:"bytes,2,opt,name=next_page_token,json=nextPageToken,proto3" json:"next_page_token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListLogEntriesResponse) Reset() {
	*x = ListLogEntriesResponse{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListLogEntriesResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListLogEntriesResponse) ProtoMessage() {}

func (x *ListLogEntriesResponse) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListLogEntriesResponse.ProtoReflect.Descriptor instead.
func (*ListLogEntriesResponse) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{5}
}

func (x *ListLogEntriesResponse) GetEntries() []*LogEntry {
	if x != nil {
		return x.Entries
	}
	return nil
}

func (x *ListLogEntriesResponse) GetNextPageToken() string {
	if x != nil {
		return x.NextPageToken
	}
	return ""
}

// The parameters to ListMonitoredResourceDescriptors
type ListMonitoredResourceDescriptorsRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Optional. The maximum number of results to return from this request.
	// Non-positive values are ignored.  The presence of `nextPageToken` in the
	// response indicates that more results might be available.
	PageSize int32 `protobuf:"varint,1,opt,name=page_size,json=pageSize,proto3" json:"page_size,omitempty"`
	// Optional. If present, then retrieve the next batch of results from the
	// preceding call to this method.  `pageToken` must be the value of
	// `nextPageToken` from the previous response.  The values of other method
	// parameters should be identical to those in the previous call.
	PageToken     string `protobuf:"bytes,2,opt,name=page_token,json=pageToken,proto3" json:"page_token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListMonitoredResourceDescriptorsRequest) Reset() {
	*x = ListMonitoredResourceDescriptorsRequest{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListMonitoredResourceDescriptorsRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListMonitoredResourceDescriptorsRequest) ProtoMessage() {}

func (x *ListMonitoredResourceDescriptorsRequest) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListMonitoredResourceDescriptorsRequest.ProtoReflect.Descriptor instead.
func (*ListMonitoredResourceDescriptorsRequest) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{6}
}

func (x *ListMonitoredResourceDescriptorsRequest) GetPageSize() int32 {
	if x != nil {
		return x.PageSize
	}
	return 0
}

func (x *ListMonitoredResourceDescriptorsRequest) GetPageToken() string {
	if x != nil {
		return x.PageToken
	}
	return ""
}

// Result returned from ListMonitoredResourceDescriptors.
type ListMonitoredResourceDescriptorsResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// A list of resource descriptors.
	ResourceDescriptors []*monitoredres.MonitoredResourceDescriptor `protobuf:"bytes,1,rep,name=resource_descriptors,json=resourceDescriptors,proto3" json:"resource_descriptors,omitempty"`
	// If there might be more results than those appearing in this response, then
	// `nextPageToken` is included.  To get the next set of results, call this
	// method again using the value of `nextPageToken` as `pageToken`.
	NextPageToken string `protobuf:"bytes,2,opt,name=next_page_token,json=nextPageToken,proto3" json:"next_page_token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListMonitoredResourceDescriptorsResponse) Reset() {
	*x = ListMonitoredResourceDescriptorsResponse{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListMonitoredResourceDescriptorsResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListMonitoredResourceDescriptorsResponse) ProtoMessage() {}

func (x *ListMonitoredResourceDescriptorsResponse) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListMonitoredResourceDescriptorsResponse.ProtoReflect.Descriptor instead.
func (*ListMonitoredResourceDescriptorsResponse) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{7}
}

func (x *ListMonitoredResourceDescriptorsResponse) GetResourceDescriptors() []*monitoredres.MonitoredResourceDescriptor {
	if x != nil {
		return x.ResourceDescriptors
	}
	return nil
}

func (x *ListMonitoredResourceDescriptorsResponse) GetNextPageToken() string {
	if x != nil {
		return x.NextPageToken
	}
	return ""
}

// The parameters to ListLogs.
type ListLogsRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Required. The resource name to list logs for:
	//
	// *  `projects/[PROJECT_ID]`
	// *  `organizations/[ORGANIZATION_ID]`
	// *  `billingAccounts/[BILLING_ACCOUNT_ID]`
	// *  `folders/[FOLDER_ID]`
	Parent string `protobuf:"bytes,1,opt,name=parent,proto3" json:"parent,omitempty"`
	// Optional. List of resource names to list logs for:
	//
	//   - `projects/[PROJECT_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `organizations/[ORGANIZATION_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `billingAccounts/[BILLING_ACCOUNT_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `folders/[FOLDER_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//
	// To support legacy queries, it could also be:
	//
	// *  `projects/[PROJECT_ID]`
	// *  `organizations/[ORGANIZATION_ID]`
	// *  `billingAccounts/[BILLING_ACCOUNT_ID]`
	// *  `folders/[FOLDER_ID]`
	//
	// The resource name in the `parent` field is added to this list.
	ResourceNames []string `protobuf:"bytes,8,rep,name=resource_names,json=resourceNames,proto3" json:"resource_names,omitempty"`
	// Optional. The maximum number of results to return from this request.
	// Non-positive values are ignored.  The presence of `nextPageToken` in the
	// response indicates that more results might be available.
	PageSize int32 `protobuf:"varint,2,opt,name=page_size,json=pageSize,proto3" json:"page_size,omitempty"`
	// Optional. If present, then retrieve the next batch of results from the
	// preceding call to this method.  `pageToken` must be the value of
	// `nextPageToken` from the previous response.  The values of other method
	// parameters should be identical to those in the previous call.
	PageToken     string `protobuf:"bytes,3,opt,name=page_token,json=pageToken,proto3" json:"page_token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListLogsRequest) Reset() {
	*x = ListLogsRequest{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListLogsRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListLogsRequest) ProtoMessage() {}

func (x *ListLogsRequest) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListLogsRequest.ProtoReflect.Descriptor instead.
func (*ListLogsRequest) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{8}
}

func (x *ListLogsRequest) GetParent() string {
	if x != nil {
		return x.Parent
	}
	return ""
}

func (x *ListLogsRequest) GetResourceNames() []string {
	if x != nil {
		return x.ResourceNames
	}
	return nil
}

func (x *ListLogsRequest) GetPageSize() int32 {
	if x != nil {
		return x.PageSize
	}
	return 0
}

func (x *ListLogsRequest) GetPageToken() string {
	if x != nil {
		return x.PageToken
	}
	return ""
}

// Result returned from ListLogs.
type ListLogsResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// A list of log names. For example,
	// `"projects/my-project/logs/syslog"` or
	// `"organizations/123/logs/cloudresourcemanager.googleapis.com%2Factivity"`.
	LogNames []string `protobuf:"bytes,3,rep,name=log_names,json=logNames,proto3" json:"log_names,omitempty"`
	// If there might be more results than those appearing in this response, then
	// `nextPageToken` is included.  To get the next set of results, call this
	// method again using the value of `nextPageToken` as `pageToken`.
	NextPageToken string `protobuf:"bytes,2,opt,name=next_page_token,json=nextPageToken,proto3" json:"next_page_token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListLogsResponse) Reset() {
	*x = ListLogsResponse{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListLogsResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListLogsResponse) ProtoMessage() {}

func (x *ListLogsResponse) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListLogsResponse.ProtoReflect.Descriptor instead.
func (*ListLogsResponse) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{9}
}

func (x *ListLogsResponse) GetLogNames() []string {
	if x != nil {
		return x.LogNames
	}
	return nil
}

func (x *ListLogsResponse) GetNextPageToken() string {
	if x != nil {
		return x.NextPageToken
	}
	return ""
}

// The parameters to `TailLogEntries`.
type TailLogEntriesRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Required. Name of a parent resource from which to retrieve log entries:
	//
	// *  `projects/[PROJECT_ID]`
	// *  `organizations/[ORGANIZATION_ID]`
	// *  `billingAccounts/[BILLING_ACCOUNT_ID]`
	// *  `folders/[FOLDER_ID]`
	//
	// May alternatively be one or more views:
	//
	//   - `projects/[PROJECT_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `organizations/[ORGANIZATION_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `billingAccounts/[BILLING_ACCOUNT_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	//   - `folders/[FOLDER_ID]/locations/[LOCATION_ID]/buckets/[BUCKET_ID]/views/[VIEW_ID]`
	ResourceNames []string `protobuf:"bytes,1,rep,name=resource_names,json=resourceNames,proto3" json:"resource_names,omitempty"`
	// Optional. Only log entries that match the filter are returned.  An empty
	// filter matches all log entries in the resources listed in `resource_names`.
	// Referencing a parent resource that is not listed in `resource_names` will
	// cause the filter to return no results. The maximum length of a filter is
	// 20,000 characters.
	Filter string `protobuf:"bytes,2,opt,name=filter,proto3" json:"filter,omitempty"`
	// Optional. The amount of time to buffer log entries at the server before
	// being returned to prevent out of order results due to late arriving log
	// entries. Valid values are between 0-60000 milliseconds. Defaults to 2000
	// milliseconds.
	BufferWindow  *durationpb.Duration `protobuf:"bytes,3,opt,name=buffer_window,json=bufferWindow,proto3" json:"buffer_window,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *TailLogEntriesRequest) Reset() {
	*x = TailLogEntriesRequest{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *TailLogEntriesRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*TailLogEntriesRequest) ProtoMessage() {}

func (x *TailLogEntriesRequest) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use TailLogEntriesRequest.ProtoReflect.Descriptor instead.
func (*TailLogEntriesRequest) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{10}
}

func (x *TailLogEntriesRequest) GetResourceNames() []string {
	if x != nil {
		return x.ResourceNames
	}
	return nil
}

func (x *TailLogEntriesRequest) GetFilter() string {
	if x != nil {
		return x.Filter
	}
	return ""
}

func (x *TailLogEntriesRequest) GetBufferWindow() *durationpb.Duration {
	if x != nil {
		return x.BufferWindow
	}
	return nil
}

// Result returned from `TailLogEntries`.
type TailLogEntriesResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// A list of log entries. Each response in the stream will order entries with
	// increasing values of `LogEntry.timestamp`. Ordering is not guaranteed
	// between separate responses.
	Entries []*LogEntry `protobuf:"bytes,1,rep,name=entries,proto3" json:"entries,omitempty"`
	// If entries that otherwise would have been included in the session were not
	// sent back to the client, counts of relevant entries omitted from the
	// session with the reason that they were not included. There will be at most
	// one of each reason per response. The counts represent the number of
	// suppressed entries since the last streamed response.
	SuppressionInfo []*TailLogEntriesResponse_SuppressionInfo `protobuf:"bytes,2,rep,name=suppression_info,json=suppressionInfo,proto3" json:"suppression_info,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *TailLogEntriesResponse) Reset() {
	*x = TailLogEntriesResponse{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *TailLogEntriesResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*TailLogEntriesResponse) ProtoMessage() {}

func (x *TailLogEntriesResponse) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use TailLogEntriesResponse.ProtoReflect.Descriptor instead.
func (*TailLogEntriesResponse) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{11}
}

func (x *TailLogEntriesResponse) GetEntries() []*LogEntry {
	if x != nil {
		return x.Entries
	}
	return nil
}

func (x *TailLogEntriesResponse) GetSuppressionInfo() []*TailLogEntriesResponse_SuppressionInfo {
	if x != nil {
		return x.SuppressionInfo
	}
	return nil
}

// Information about entries that were omitted from the session.
type TailLogEntriesResponse_SuppressionInfo struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// The reason that entries were omitted from the session.
	Reason TailLogEntriesResponse_SuppressionInfo_Reason `protobuf:"varint,1,opt,name=reason,proto3,enum=google.logging.v2.TailLogEntriesResponse_SuppressionInfo_Reason" json:"reason,omitempty"`
	// A lower bound on the count of entries omitted due to `reason`.
	SuppressedCount int32 `protobuf:"varint,2,opt,name=suppressed_count,json=suppressedCount,proto3" json:"suppressed_count,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *TailLogEntriesResponse_SuppressionInfo) Reset() {
	*x = TailLogEntriesResponse_SuppressionInfo{}
	mi := &file_google_logging_v2_logging_proto_msgTypes[14]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *TailLogEntriesResponse_SuppressionInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*TailLogEntriesResponse_SuppressionInfo) ProtoMessage() {}

func (x *TailLogEntriesResponse_SuppressionInfo) ProtoReflect() protoreflect.Message {
	mi := &file_google_logging_v2_logging_proto_msgTypes[14]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use TailLogEntriesResponse_SuppressionInfo.ProtoReflect.Descriptor instead.
func (*TailLogEntriesResponse_SuppressionInfo) Descriptor() ([]byte, []int) {
	return file_google_logging_v2_logging_proto_rawDescGZIP(), []int{11, 0}
}

func (x *TailLogEntriesResponse_SuppressionInfo) GetReason() TailLogEntriesResponse_SuppressionInfo_Reason {
	if x != nil {
		return x.Reason
	}
	return TailLogEntriesResponse_SuppressionInfo_REASON_UNSPECIFIED
}

func (x *TailLogEntriesResponse_SuppressionInfo) GetSuppressedCount() int32 {
	if x != nil {
		return x.SuppressedCount
	}
	return 0
}

var File_google_logging_v2_logging_proto protoreflect.FileDescriptor

const file_google_logging_v2_logging_proto_rawDesc = "" +
	"\n" +
	"\x1fgoogle/logging/v2/logging.proto\x12\x11google.logging.v2\x1a\x1cgoogle/api/annotations.proto\x1a\x17google/api/client.proto\x1a\x1fgoogle/api/field_behavior.proto\x1a#google/api/monitored_resource.proto\x1a\x19google/api/resource.proto\x1a!google/logging/v2/log_entry.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x17google/rpc/status.proto\"Q\n" +
	"\x10DeleteLogRequest\x12=\n" +
	"\blog_name\x18\x01 \x01(\tB\"\xe0A\x02\xfaA\x1c\n" +
	"\x1alogging.googleapis.com/LogR\alogName\"\xae\x03\n" +
	"\x16WriteLogEntriesRequest\x12=\n" +
	"\blog_name\x18\x01 \x01(\tB\"\xe0A\x01\xfaA\x1c\n" +
	"\x1alogging.googleapis.com/LogR\alogName\x12>\n" +
	"\bresource\x18\x02 \x01(\v2\x1d.google.api.MonitoredResourceB\x03\xe0A\x01R\bresource\x12R\n" +
	"\x06labels\x18\x03 \x03(\v25.google.logging.v2.WriteLogEntriesRequest.LabelsEntryB\x03\xe0A\x01R\x06labels\x12:\n" +
	"\aentries\x18\x04 \x03(\v2\x1b.google.logging.v2.LogEntryB\x03\xe0A\x02R\aentries\x12,\n" +
	"\x0fpartial_success\x18\x05 \x01(\bB\x03\xe0A\x01R\x0epartialSuccess\x12\x1c\n" +
	"\adry_run\x18\x06 \x01(\bB\x03\xe0A\x01R\x06dryRun\x1a9\n" +
	"\vLabelsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01\"\x19\n" +
	"\x17WriteLogEntriesResponse\"\xe4\x01\n" +
	"\x1cWriteLogEntriesPartialErrors\x12m\n" +
	"\x10log_entry_errors\x18\x01 \x03(\v2C.google.logging.v2.WriteLogEntriesPartialErrors.LogEntryErrorsEntryR\x0elogEntryErrors\x1aU\n" +
	"\x13LogEntryErrorsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\x05R\x03key\x12(\n" +
	"\x05value\x18\x02 \x01(\v2\x12.google.rpc.StatusR\x05value:\x028\x01\"\xe5\x01\n" +
	"\x15ListLogEntriesRequest\x12I\n" +
	"\x0eresource_names\x18\b \x03(\tB\"\xe0A\x02\xfaA\x1c\x12\x1alogging.googleapis.com/LogR\rresourceNames\x12\x1b\n" +
	"\x06filter\x18\x02 \x01(\tB\x03\xe0A\x01R\x06filter\x12\x1e\n" +
	"\border_by\x18\x03 \x01(\tB\x03\xe0A\x01R\aorderBy\x12 \n" +
	"\tpage_size\x18\x04 \x01(\x05B\x03\xe0A\x01R\bpageSize\x12\"\n" +
	"\n" +
	"page_token\x18\x05 \x01(\tB\x03\xe0A\x01R\tpageToken\"w\n" +
	"\x16ListLogEntriesResponse\x125\n" +
	"\aentries\x18\x01 \x03(\v2\x1b.google.logging.v2.LogEntryR\aentries\x12&\n" +
	"\x0fnext_page_token\x18\x02 \x01(\tR\rnextPageToken\"o\n" +
	"'ListMonitoredResourceDescriptorsRequest\x12 \n" +
	"\tpage_size\x18\x01 \x01(\x05B\x03\xe0A\x01R\bpageSize\x12\"\n" +
	"\n" +
	"page_token\x18\x02 \x01(\tB\x03\xe0A\x01R\tpageToken\"\xae\x01\n" +
	"(ListMonitoredResourceDescriptorsResponse\x12Z\n" +
	"\x14resource_descriptors\x18\x01 \x03(\v2'.google.api.MonitoredResourceDescriptorR\x13resourceDescriptors\x12&\n" +
	"\x0fnext_page_token\x18\x02 \x01(\tR\rnextPageToken\"\xde\x01\n" +
	"\x0fListLogsRequest\x12:\n" +
	"\x06parent\x18\x01 \x01(\tB\"\xe0A\x02\xfaA\x1c\x12\x1alogging.googleapis.com/LogR\x06parent\x12I\n" +
	"\x0eresource_names\x18\b \x03(\tB\"\xe0A\x01\xfaA\x1c\x12\x1alogging.googleapis.com/LogR\rresourceNames\x12 \n" +
	"\tpage_size\x18\x02 \x01(\x05B\x03\xe0A\x01R\bpageSize\x12\"\n" +
	"\n" +
	"page_token\x18\x03 \x01(\tB\x03\xe0A\x01R\tpageToken\"W\n" +
	"\x10ListLogsResponse\x12\x1b\n" +
	"\tlog_names\x18\x03 \x03(\tR\blogNames\x12&\n" +
	"\x0fnext_page_token\x18\x02 \x01(\tR\rnextPageToken\"\xa5\x01\n" +
	"\x15TailLogEntriesRequest\x12*\n" +
	"\x0eresource_names\x18\x01 \x03(\tB\x03\xe0A\x02R\rresourceNames\x12\x1b\n" +
	"\x06filter\x18\x02 \x01(\tB\x03\xe0A\x01R\x06filter\x12C\n" +
	"\rbuffer_window\x18\x03 \x01(\v2\x19.google.protobuf.DurationB\x03\xe0A\x01R\fbufferWindow\"\x92\x03\n" +
	"\x16TailLogEntriesResponse\x125\n" +
	"\aentries\x18\x01 \x03(\v2\x1b.google.logging.v2.LogEntryR\aentries\x12d\n" +
	"\x10suppression_info\x18\x02 \x03(\v29.google.logging.v2.TailLogEntriesResponse.SuppressionInfoR\x0fsuppressionInfo\x1a\xda\x01\n" +
	"\x0fSuppressionInfo\x12X\n" +
	"\x06reason\x18\x01 \x01(\x0e2@.google.logging.v2.TailLogEntriesResponse.SuppressionInfo.ReasonR\x06reason\x12)\n" +
	"\x10suppressed_count\x18\x02 \x01(\x05R\x0fsuppressedCount\"B\n" +
	"\x06Reason\x12\x16\n" +
	"\x12REASON_UNSPECIFIED\x10\x00\x12\x0e\n" +
	"\n" +
	"RATE_LIMIT\x10\x01\x12\x10\n" +
	"\fNOT_CONSUMED\x10\x022\xe9\r\n" +
	"\x10LoggingServiceV2\x12\x93\x02\n" +
	"\tDeleteLog\x12#.google.logging.v2.DeleteLogRequest\x1a\x16.google.protobuf.Empty\"\xc8\x01\xdaA\blog_name\x82\xd3\xe4\x93\x02\xb6\x01Z\x1b*\x19/v2/{log_name=*/*/logs/*}Z'*%/v2/{log_name=organizations/*/logs/*}Z!*\x1f/v2/{log_name=folders/*/logs/*}Z)*'/v2/{log_name=billingAccounts/*/logs/*}* /v2/{log_name=projects/*/logs/*}\x12\xa9\x01\n" +
	"\x0fWriteLogEntries\x12).google.logging.v2.WriteLogEntriesRequest\x1a*.google.logging.v2.WriteLogEntriesResponse\"?\xdaA log_name,resource,labels,entries\x82\xd3\xe4\x93\x02\x16:\x01*\"\x11/v2/entries:write\x12\xa3\x01\n" +
	"\x0eListLogEntries\x12(.google.logging.v2.ListLogEntriesRequest\x1a).google.logging.v2.ListLogEntriesResponse\"<\xdaA\x1eresource_names,filter,order_by\x82\xd3\xe4\x93\x02\x15:\x01*\"\x10/v2/entries:list\x12\xc5\x01\n" +
	" ListMonitoredResourceDescriptors\x12:.google.logging.v2.ListMonitoredResourceDescriptorsRequest\x1a;.google.logging.v2.ListMonitoredResourceDescriptorsResponse\"(\x82\xd3\xe4\x93\x02\"\x12 /v2/monitoredResourceDescriptors\x12\x8b\x04\n" +
	"\bListLogs\x12\".google.logging.v2.ListLogsRequest\x1a#.google.logging.v2.ListLogsResponse\"\xb5\x03\xdaA\x06parent\x82\xd3\xe4\x93\x02\xa5\x03Z\x1e\x12\x1c/v2/{parent=projects/*}/logsZ#\x12!/v2/{parent=organizations/*}/logsZ\x1d\x12\x1b/v2/{parent=folders/*}/logsZ%\x12#/v2/{parent=billingAccounts/*}/logsZ<\x12:/v2/{parent=projects/*/locations/*/buckets/*/views/*}/logsZA\x12?/v2/{parent=organizations/*/locations/*/buckets/*/views/*}/logsZ;\x129/v2/{parent=folders/*/locations/*/buckets/*/views/*}/logsZC\x12A/v2/{parent=billingAccounts/*/locations/*/buckets/*/views/*}/logs\x12\x15/v2/{parent=*/*}/logs\x12\x86\x01\n" +
	"\x0eTailLogEntries\x12(.google.logging.v2.TailLogEntriesRequest\x1a).google.logging.v2.TailLogEntriesResponse\"\x1b\x82\xd3\xe4\x93\x02\x15:\x01*\"\x10/v2/entries:tail(\x010\x01\x1a\x8d\x02\xcaA\x16logging.googleapis.com\xd2A\xf0\x01https://www.googleapis.com/auth/cloud-platform,https://www.googleapis.com/auth/cloud-platform.read-only,https://www.googleapis.com/auth/logging.admin,https://www.googleapis.com/auth/logging.read,https://www.googleapis.com/auth/logging.writeB\xb2\x01\n" +
	"\x15com.google.logging.v2B\fLoggingProtoP\x01Z5cloud.google.com/go/logging/apiv2/loggingpb;loggingpb\xf8\x01\x01\xaa\x02\x17Google.Cloud.Logging.V2\xca\x02\x17Google\\Cloud\\Logging\\V2\xea\x02\x1aGoogle::Cloud::Logging::V2b\x06proto3"

var (
	file_google_logging_v2_logging_proto_rawDescOnce sync.Once
	file_google_logging_v2_logging_proto_rawDescData []byte
)

func file_google_logging_v2_logging_proto_rawDescGZIP() []byte {
	file_google_logging_v2_logging_proto_rawDescOnce.Do(func() {
		file_google_logging_v2_logging_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_google_logging_v2_logging_proto_rawDesc), len(file_google_logging_v2_logging_proto_rawDesc)))
	})
	return file_google_logging_v2_logging_proto_rawDescData
}

var file_google_logging_v2_logging_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_google_logging_v2_logging_proto_msgTypes = make([]protoimpl.MessageInfo, 15)
var file_google_logging_v2_logging_proto_goTypes = []any{
	(TailLogEntriesResponse_SuppressionInfo_Reason)(0), // 0: google.logging.v2.TailLogEntriesResponse.SuppressionInfo.Reason
	(*DeleteLogRequest)(nil),                           // 1: google.logging.v2.DeleteLogRequest
	(*WriteLogEntriesRequest)(nil),                     // 2: google.logging.v2.WriteLogEntriesRequest
	(*WriteLogEntriesResponse)(nil),                    // 3: google.logging.v2.WriteLogEntriesResponse
	(*WriteLogEntriesPartialErrors)(nil),               // 4: google.logging.v2.WriteLogEntriesPartialErrors
	(*ListLogEntriesRequest)(nil),                      // 5: google.logging.v2.ListLogEntriesRequest
	(*ListLogEntriesResponse)(nil),                     // 6: google.logging.v2.ListLogEntriesResponse
	(*ListMonitoredResourceDescriptorsRequest)(nil),    // 7: google.logging.v2.ListMonitoredResourceDescriptorsRequest
	(*ListMonitoredResourceDescriptorsResponse)(nil),   // 8: google.logging.v2.ListMonitoredResourceDescriptorsResponse
	(*ListLogsRequest)(nil),                            // 9: google.logging.v2.ListLogsRequest
	(*ListLogsResponse)(nil),                           // 10: google.logging.v2.ListLogsResponse
	(*TailLogEntriesRequest)(nil),                      // 11: google.logging.v2.TailLogEntriesRequest
	(*TailLogEntriesResponse)(nil),                     // 12: google.logging.v2.TailLogEntriesResponse
	nil,                                                // 13: google.logging.v2.WriteLogEntriesRequest.LabelsEntry
	nil,                                                // 14: google.logging.v2.WriteLogEntriesPartialErrors.LogEntryErrorsEntry
	(*TailLogEntriesResponse_SuppressionInfo)(nil),     // 15: google.logging.v2.TailLogEntriesResponse.SuppressionInfo
	(*monitoredres.MonitoredResource)(nil),             // 16: google.api.MonitoredResource
	(*LogEntry)(nil),                                   // 17: google.logging.v2.LogEntry
	(*monitoredres.MonitoredResourceDescriptor)(nil),   // 18: google.api.MonitoredResourceDescriptor
	(*durationpb.Duration)(nil),                        // 19: google.protobuf.Duration
	(*status.Status)(nil),                              // 20: google.rpc.Status
	(*emptypb.Empty)(nil),                              // 21: google.protobuf.Empty
}
var file_google_logging_v2_logging_proto_depIdxs = []int32{
	16, // 0: google.logging.v2.WriteLogEntriesRequest.resource:type_name -> google.api.MonitoredResource
	13, // 1: google.logging.v2.WriteLogEntriesRequest.labels:type_name -> google.logging.v2.WriteLogEntriesRequest.LabelsEntry
	17, // 2: google.logging.v2.WriteLogEntriesRequest.entries:type_name -> google.logging.v2.LogEntry
	14, // 3: google.logging.v2.WriteLogEntriesPartialErrors.log_entry_errors:type_name -> google.logging.v2.WriteLogEntriesPartialErrors.LogEntryErrorsEntry
	17, // 4: google.logging.v2.ListLogEntriesResponse.entries:type_name -> google.logging.v2.LogEntry
	18, // 5: google.logging.v2.ListMonitoredResourceDescriptorsResponse.resource_descriptors:type_name -> google.api.MonitoredResourceDescriptor
	19, // 6: google.logging.v2.TailLogEntriesRequest.buffer_window:type_name -> google.protobuf.Duration
	17, // 7: google.logging.v2.TailLogEntriesResponse.entries:type_name -> google.logging.v2.LogEntry
	15, // 8: google.logging.v2.TailLogEntriesResponse.suppression_info:type_name -> google.logging.v2.TailLogEntriesResponse.SuppressionInfo
	20, // 9: google.logging.v2.WriteLogEntriesPartialErrors.LogEntryErrorsEntry.value:type_name -> google.rpc.Status
	0,  // 10: google.logging.v2.TailLogEntriesResponse.SuppressionInfo.reason:type_name -> google.logging.v2.TailLogEntriesResponse.SuppressionInfo.Reason
	1,  // 11: google.logging.v2.LoggingServiceV2.DeleteLog:input_type -> google.logging.v2.DeleteLogRequest
	2,  // 12: google.logging.v2.LoggingServiceV2.WriteLogEntries:input_type -> google.logging.v2.WriteLogEntriesRequest
	5,  // 13: google.logging.v2.LoggingServiceV2.ListLogEntries:input_type -> google.logging.v2.ListLogEntriesRequest
	7,  // 14: google.logging.v2.LoggingServiceV2.ListMonitoredResourceDescriptors:input_type -> google.logging.v2.ListMonitoredResourceDescriptorsRequest
	9,  // 15: google.logging.v2.LoggingServiceV2.ListLogs:input_type -> google.logging.v2.ListLogsRequest
	11, // 16: google.logging.v2.LoggingServiceV2.TailLogEntries:input_type -> google.logging.v2.TailLogEntriesRequest
	21, // 17: google.logging.v2.LoggingServiceV2.DeleteLog:output_type -> google.protobuf.Empty
	3,  // 18: google.logging.v2.LoggingServiceV2.WriteLogEntries:output_type -> google.logging.v2.WriteLogEntriesResponse
	6,  // 19: google.logging.v2.LoggingServiceV2.ListLogEntries:output_type -> google.logging.v2.ListLogEntriesResponse
	8,  // 20: google.logging.v2.LoggingServiceV2.ListMonitoredResourceDescriptors:output_type -> google.logging.v2.ListMonitoredResourceDescriptorsResponse
	10, // 21: google.logging.v2.LoggingServiceV2.ListLogs:output_type -> google.logging.v2.ListLogsResponse
	12, // 22: google.logging.v2.LoggingServiceV2.TailLogEntries:output_type -> google.logging.v2.TailLogEntriesResponse
	17, // [17:23] is the sub-list for method output_type
	11, // [11:17] is the sub-list for method input_type
	11, // [11:11] is the sub-list for extension type_name
	11, // [11:11] is the sub-list for extension extendee
	0,  // [0:11] is the sub-list for field type_name
}

func init() { file_google_logging_v2_logging_proto_init() }
func file_google_logging_v2_logging_proto_init() {
	if File_google_logging_v2_logging_proto != nil {
		return
	}
	file_google_logging_v2_log_entry_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_google_logging_v2_logging_proto_rawDesc), len(file_google_logging_v2_logging_proto_rawDesc)),
			NumEnums:      1,
			NumMessages:   15,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_google_logging_v2_logging_proto_goTypes,
		DependencyIndexes: file_google_logging_v2_logging_proto_depIdxs,
		EnumInfos:         file_google_logging_v2_logging_proto_enumTypes,
		MessageInfos:      file_google_logging_v2_logging_proto_msgTypes,
	}.Build()
	File_google_logging_v2_logging_proto = out.File
	file_google_logging_v2_logging_proto_goTypes = nil
	file_google_logging_v2_logging_proto_depIdxs = nil
}
