// Code generated by smithy-go-codegen DO NOT EDIT.

package cloudwatchlogs

import (
	"context"
	"fmt"
	awsmiddleware "github.com/aws/aws-sdk-go-v2/aws/middleware"
	"github.com/aws/smithy-go/middleware"
	smithyhttp "github.com/aws/smithy-go/transport/http"
)

// Creates a lookup table by uploading CSV data. You can use lookup tables to
// enrich log data in CloudWatch Logs Insights queries with reference data such as
// user details, application names, or error descriptions.
//
// The table name must be unique within your account and Region. The CSV content
// must include a header row with column names, use UTF-8 encoding, and not exceed
// 10 MB.
func (c *Client) CreateLookupTable(ctx context.Context, params *CreateLookupTableInput, optFns ...func(*Options)) (*CreateLookupTableOutput, error) {
	if params == nil {
		params = &CreateLookupTableInput{}
	}

	result, metadata, err := c.invokeOperation(ctx, "CreateLookupTable", params, optFns, c.addOperationCreateLookupTableMiddlewares)
	if err != nil {
		return nil, err
	}

	out := result.(*CreateLookupTableOutput)
	out.ResultMetadata = metadata
	return out, nil
}

type CreateLookupTableInput struct {

	// The name of the lookup table. The name must be unique within your account and
	// Region. The name can contain only alphanumeric characters and underscores, and
	// can be up to 256 characters long.
	//
	// This member is required.
	LookupTableName *string

	// The CSV content of the lookup table. The first row must be a header row with
	// column names. The content must use UTF-8 encoding and not exceed 10 MB.
	//
	// This member is required.
	TableBody *string

	// A description of the lookup table. The description can be up to 1024 characters
	// long.
	Description *string

	// The ARN of the KMS key to use to encrypt the lookup table data. If you don't
	// specify a key, the data is encrypted with an Amazon Web Services-owned key.
	KmsKeyId *string

	// A list of key-value pairs to associate with the lookup table. You can associate
	// as many as 50 tags with a lookup table. Tags can help you organize and
	// categorize your resources.
	Tags map[string]string

	noSmithyDocumentSerde
}

type CreateLookupTableOutput struct {

	// The time when the lookup table was created, expressed as the number of
	// milliseconds after Jan 1, 1970 00:00:00 UTC .
	CreatedAt *int64

	// The ARN of the lookup table that was created.
	LookupTableArn *string

	// Metadata pertaining to the operation's result.
	ResultMetadata middleware.Metadata

	noSmithyDocumentSerde
}

func (c *Client) addOperationCreateLookupTableMiddlewares(stack *middleware.Stack, options Options) (err error) {
	if err := stack.Serialize.Add(&setOperationInputMiddleware{}, middleware.After); err != nil {
		return err
	}
	err = stack.Serialize.Add(&awsAwsjson11_serializeOpCreateLookupTable{}, middleware.After)
	if err != nil {
		return err
	}
	err = stack.Deserialize.Add(&awsAwsjson11_deserializeOpCreateLookupTable{}, middleware.After)
	if err != nil {
		return err
	}
	if err := addProtocolFinalizerMiddlewares(stack, options, "CreateLookupTable"); err != nil {
		return fmt.Errorf("add protocol finalizers: %v", err)
	}

	if err = addlegacyEndpointContextSetter(stack, options); err != nil {
		return err
	}
	if err = addSetLoggerMiddleware(stack, options); err != nil {
		return err
	}
	if err = addClientRequestID(stack); err != nil {
		return err
	}
	if err = addComputeContentLength(stack); err != nil {
		return err
	}
	if err = addResolveEndpointMiddleware(stack, options); err != nil {
		return err
	}
	if err = addComputePayloadSHA256(stack); err != nil {
		return err
	}
	if err = addRetry(stack, options, c); err != nil {
		return err
	}
	if err = addRawResponseToMetadata(stack); err != nil {
		return err
	}
	if err = addRecordResponseTiming(stack); err != nil {
		return err
	}
	if err = addSpanRetryLoop(stack, options); err != nil {
		return err
	}
	if err = addClientUserAgent(stack, options); err != nil {
		return err
	}
	if err = smithyhttp.AddErrorCloseResponseBodyMiddleware(stack); err != nil {
		return err
	}
	if err = smithyhttp.AddCloseResponseBodyMiddleware(stack); err != nil {
		return err
	}
	if err = addSetLegacyContextSigningOptionsMiddleware(stack); err != nil {
		return err
	}
	if err = addUserAgentRetryMode(stack, options); err != nil {
		return err
	}
	if err = addCredentialSource(stack, options); err != nil {
		return err
	}
	if err = addOpCreateLookupTableValidationMiddleware(stack); err != nil {
		return err
	}
	if err = stack.Initialize.Add(newServiceMetadataMiddleware_opCreateLookupTable(options.Region), middleware.Before); err != nil {
		return err
	}
	if err = addRecursionDetection(stack); err != nil {
		return err
	}
	if err = addRequestIDRetrieverMiddleware(stack); err != nil {
		return err
	}
	if err = addResponseErrorMiddleware(stack); err != nil {
		return err
	}
	if err = addRequestResponseLogging(stack, options); err != nil {
		return err
	}
	if err = addDisableHTTPSMiddleware(stack, options); err != nil {
		return err
	}
	if err = addInterceptBeforeRetryLoop(stack, options); err != nil {
		return err
	}
	if err = addInterceptAttempt(stack, options); err != nil {
		return err
	}
	if err = addInterceptors(stack, options); err != nil {
		return err
	}
	return nil
}

func newServiceMetadataMiddleware_opCreateLookupTable(region string) *awsmiddleware.RegisterServiceMetadata {
	return &awsmiddleware.RegisterServiceMetadata{
		Region:        region,
		ServiceID:     ServiceID,
		OperationName: "CreateLookupTable",
	}
}
