package strslice

// StrSlice represents a string or an array of strings.
//
// Deprecated: this type was used for compatibility with deprecated API versions. Use []string instead.
//
//go:fix inline
type StrSlice = []string
