// Autogenerated code; DO NOT EDIT.

/*
 * Schema Open API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

// Configuration of a virtual machine, used during its creation to set up and/or use resources.
type VirtualMachine struct {
	Version *Version `json:"Version,omitempty"`
	// When set to true, the virtual machine will treat a reset as a stop, releasing resources and cleaning up state.
	StopOnReset     bool             `json:"StopOnReset,omitempty"`
	Chipset         *Chipset         `json:"Chipset,omitempty"`
	ComputeTopology *Topology        `json:"ComputeTopology,omitempty"`
	Devices         *Devices         `json:"Devices,omitempty"`
	GuestState      *GuestState      `json:"GuestState,omitempty"`
	RestoreState    *RestoreState    `json:"RestoreState,omitempty"`
	RegistryChanges *RegistryChanges `json:"RegistryChanges,omitempty"`
	StorageQoS      *StorageQoS      `json:"StorageQoS,omitempty"`
	DebugOptions    *DebugOptions    `json:"DebugOptions,omitempty"`
	GuestConnection *GuestConnection `json:"GuestConnection,omitempty"`
	SecuritySettings  *SecuritySettings `json:"SecuritySettings,omitempty"`
}
