// Copyright The Moby Authors.
// SPDX-License-Identifier: Apache-2.0

//go:build linux

package apparmor

// NOTE: This profile is replicated in containerd and libpod. If you make a
//       change to this profile, please make follow-up PRs to those projects so
//       that these rules can be synchronised (because any issue with this
//       profile will likely affect libpod and containerd).

// baseTemplate defines the default apparmor profile for containers.
//
// It explicitly sets the AppArmor ABI to 3.0. In AppArmor ABI higher than 4.0,
// "network" no longer includes "network unix", resulting in access to unix sockets
// being denied. We use ABI 3.0 to account for some LTS distros that do not
// yet support ABI 4.0.
//
// See https://gitlab.com/apparmor/apparmor/-/issues/561
// And https://github.com/containerd/containerd/issues/12726
const baseTemplate = `# profile generated by github.com/moby/profiles/apparmor.
{{if .Abi}}abi <{{.Abi}}>,
{{- end}}
{{range $value := .Imports}}
{{$value}}
{{- end}}

profile "{{.Name}}" flags=(attach_disconnected,mediate_deleted) {
{{- range $value := .InnerImports}}
  {{$value}}
{{- end}}{{if .InnerImports}}
{{end}}
  network,
  # Disallow AF_ALG (Linux kernel crypto API); see https://copy.fail/
  deny network alg,
  capability,
  file,
  umount,
  # Host (privileged) processes may send signals to container processes.
  signal (receive) peer=unconfined,
  # runc may send signals to container processes (for "docker stop").
  signal (receive) peer=runc,
  # crun may send signals to container processes (for "docker stop" when used with crun OCI runtime).
  signal (receive) peer=crun,
  # dockerd may send signals to container processes (for "docker kill").
  signal (receive) peer="{{.DaemonProfile}}",
  # Container processes may send signals amongst themselves.
  signal (send,receive) peer="{{.Name}}",

  deny @{PROC}/* w,   # deny write for all files directly in /proc (not in a subdir)
  # deny write to files not in /proc/<number>/** or /proc/sys/**
  deny @{PROC}/{[^1-9/],[^1-9/][^0-9/],[^1-9s/][^0-9y/][^0-9s/],[^1-9/][^0-9/][^0-9/][^0-9/]*}/** w,
  deny @{PROC}/sys/[^k]** w,  # deny /proc/sys except /proc/sys/k* (effectively /proc/sys/kernel)
  deny @{PROC}/sys/kernel/{?,??,[^s][^h][^m]**} w,  # deny everything except shm* in /proc/sys/kernel/
  deny @{PROC}/sysrq-trigger rwklx,
  deny @{PROC}/kcore rwklx,

  deny mount,

  deny /sys/[^f]*/** wklx,
  deny /sys/f[^s]*/** wklx,
  deny /sys/fs/[^c]*/** wklx,
  deny /sys/fs/c[^g]*/** wklx,
  deny /sys/fs/cg[^r]*/** wklx,
  deny /sys/firmware/** rwklx,
  deny /sys/devices/virtual/powercap/** rwklx,
  deny /sys/kernel/security/** rwklx,

  # allow processes within the container to trace each other,
  # provided all other LSM and yama setting allow it.
  ptrace (trace,tracedby,read,readby) peer="{{.Name}}",
}
`
